from __future__ import print_function
import sys, os
import ctypes
import binascii
import pprint
import struct
import csv

DEBUG = False

FORMAT_MODE = 'current'
VALID_FORMAT_MODES =('current', 'partialpressure', 'concentration', 'torr', 'mbar', 'pa')

# constants for partial pressure calculations and unit conversions
HYDROGEN_FF = 0.98
HYDROGEN_XF = XF = 0.44
SENSITIVITY_CONSTANT = 1
FRAG_FACTOR_CONSTANT = .94
ION_PROB_CONSTANT = 1

class ChannelDescriptor(ctypes.BigEndianStructure):
    mode = ""
    _fields_ = [
        ('raw_mode',    ctypes.c_char * 32),
        ('enabled',     ctypes.c_uint16),
        ('start_mass',  ctypes.c_uint16),
        ('stop_mass',   ctypes.c_uint16),
        ('ppamu',       ctypes.c_uint16),
        ('extra',       ctypes.c_uint16),
        ('dwell',       ctypes.c_uint16),
        ('data_format',     ctypes.c_uint16),
        ('pts_per_chan',    ctypes.c_uint16),
    ]

    def load(self, f):
        size = ctypes.sizeof(self)
        bytes = f.read(size)
        if len(bytes) == 0:
            return False
        ctypes.memmove(ctypes.addressof(self), bytes, len(bytes))
        self.mode = self.raw_mode.decode()
        return True

    def dump(self):
        print("mode:         %s" % self.mode)
        print("enabled:      %d" % self.enabled)
        print("start_mass:   %d" % self.start_mass)
        print("stop_mass:    %d" % self.stop_mass)
        print("ppamu:        %d" % self.ppamu)
        print("dwell:        %d" % self.dwell)
        print("extra:        %d" % self.extra)
        print("data_format:  %d" % self.data_format)
        print("pts_per_chan: %d" % self.pts_per_chan)

class GroupHeader(ctypes.BigEndianStructure):
    _fields_ = [
        ('start_scan',      ctypes.c_uint32),
        ('num_scans',       ctypes.c_uint32),
        ('pts_per_scan',    ctypes.c_uint32),
        ('error_dropped',   ctypes.c_uint32),
        ('error_overflow',  ctypes.c_uint32),
        ('reserved',        ctypes.c_uint32),
    ]

    def load(self, f):
        size = ctypes.sizeof(self)
        bytes = f.read(size)
        if len(bytes) == 0:
            return False
        ctypes.memmove(ctypes.addressof(self), bytes, len(bytes))
        return True

    def dump(self):
        print("start_scan:     %d" % self.start_scan)
        print("num_scans:      %d" % self.num_scans)
        print("pts_per_scan:   %d" % self.pts_per_scan)
        print("error_dropped:  %d" % self.error_dropped)
        print("error_overflow: %d" % self.error_overflow)
        print("reserved:       %d" % self.reserved)

    def meta_row(self):
        print("META,%d,%d,%d,%d,%d,%d" % (
            self.start_scan, self.num_scans, self.pts_per_scan,
            self.error_dropped, self.error_overflow, self.reserved
        ))

def read_file_header(f):
    chan = ChannelDescriptor()
    data = f.read(40)
    trim = data[-4:]
    num_channels,  = struct.unpack(">L", trim);
    channels = []
    if DEBUG:
        print("Found %d channels" % num_channels)
    pts_per_scan = 0
    scan_format  = [">"]
    header_parse = []
    for idx in range(num_channels):
        if DEBUG:
            print("")
            print("Channel", idx+1)
        chan.load(f)
        if DEBUG:
            chan.dump()

        if chan.mode == "Sweep":
            ppamu = chan.ppamu
            mode = chan.mode
            start = chan.start_mass
            pts = chan.pts_per_chan
            num = 0
            for i in range(pts):
                num = float(start + (100/ppamu)*i) / 100
                head = mode + ": " + str(num)
                header_parse.append(head)
        elif chan.mode == "Single":
            ppamu = chan.ppamu
            mode = chan.mode
            start = chan.start_mass
            num = float(start) / 100
            head = mode + ": " + str(num)
            header_parse.append(head)
        else:
            header_parse.append(chan.mode)

        if chan.data_format == 1: 
            pts_per_scan += chan.pts_per_chan
            scan_format.append("f" * chan.pts_per_chan)
        elif chan.data_format == 2:
            pts_per_scan += chan.pts_per_chan
            scan_format.append("L" * chan.pts_per_chan)

    unpack_format = "".join(scan_format)
    if DEBUG:
        print("Points Per Scan:", pts_per_scan)
        print("Unpack Format:", unpack_format)

    return header_parse, pts_per_scan, unpack_format


def group_iterator(f):
    while True:
        header = GroupHeader()
        if header.load(f):
            yield header
        else:
            return

def scan_iterator(f, header, unpack_format):
    start = header.start_scan
    end = start+header.num_scans
    pps = header.pts_per_scan

    for n in range(start, end):
        data = f.read(4*pps)
        scan =  (n, ) + struct.unpack(unpack_format, data)
        yield scan
        
def check_mass_field(field):
    """
    Return the mass of the field if it is a Sweep field or Single field.
    Otherwise, return None.
    """
    mass = None
    if 'Sweep: ' in field or 'Single: ' in field:
        mass = float(field.split(': ')[1])

    return mass
    
def current_to_pp(current):
    """
    Convert current reading to partial pressure.
    """
    return current / (HYDROGEN_FF * HYDROGEN_XF * SENSITIVITY_CONSTANT)

def write_output_file(fname):
    out_Name = fname[0:-4] + '.csv'
    with open(out_Name, mode='w') as output_file:
        header = GroupHeader()
        headers = []
        writer = csv.writer(output_file, lineterminator='\n')
        console = csv.writer(sys.stdout, lineterminator='\n')
        f = open(fname, 'rb')

        header_parse, pts_per_scan, unpack_format = read_file_header(f)

        header_parse.insert(0,"Scan Number")

        # remove non-whole number sweep fields from header
        t_header = []
        for field in header_parse:
            mass = check_mass_field(field)
            if mass is not None and not mass.is_integer():
                continue
            t_header.append(field)

        writer.writerow(t_header)

        for header in group_iterator(f):
            if DEBUG:
                header.meta_row()
            if header.pts_per_scan != pts_per_scan:
                raise Exception("Points Per Scan Mismatch [%d, %d]" % (header.pts_per_scan, pts_per_scan))

            total_pressure_idx = header_parse.index('TotalPressure')

            for scan in scan_iterator(f, header, unpack_format):
                # remove values corresponding to non-whole number masses
                t_scan = ()
                for field_idx in range(len(header_parse)):
                    mass = check_mass_field(header_parse[field_idx])
                    if mass is not None and not mass.is_integer():
                        continue

                    value = scan[field_idx]
                    if mass is not None:
                        if FORMAT_MODE == 'partialpressure':
                            # convert current to partial pressure
                            value = current_to_pp(abs(value))
                        elif FORMAT_MODE == 'concentration':
                            # convert current to concentration in ppm
                            pp = current_to_pp(abs(value))
                            tp = scan[total_pressure_idx]
                            value = (pp / tp) * 1000000
                        elif FORMAT_MODE == 'torr':
                            value =  value * (1 / (SENSITIVITY_CONSTANT * FRAG_FACTOR_CONSTANT * ION_PROB_CONSTANT))
                        elif FORMAT_MODE == 'mbar':
                            value =  value * (1.333222 / (SENSITIVITY_CONSTANT * FRAG_FACTOR_CONSTANT * ION_PROB_CONSTANT))
                        elif FORMAT_MODE == 'pa':
                            value =  value * (133.3222 / (SENSITIVITY_CONSTANT * FRAG_FACTOR_CONSTANT * ION_PROB_CONSTANT))


                    t_scan = t_scan + (value, )
                writer.writerow(t_scan)
                if DEBUG:
                    console.writerow(scan)




if __name__ == "__main__":
    if "-v" in sys.argv:
        DEBUG = True
        sys.argv.remove("-v") 
    elif "--help" in sys.argv:
        print("Usage: python msd_to_csv.py <.msd to be converted to .csv> [options]\n")
        print("Options:")
        print("  -v         - enable debug mode")
        print("  -f <mode>  - output format mode: " + str(VALID_FORMAT_MODES)[1:-1])
        print("  -s <value> - sensitivity constant")
        print("  -p <value> - ionization probability")
        print("  -g <value> - frag factor")
        sys.exit()
    else:
        DEBUG = False

    if '-f' in sys.argv:
        # data format specifier
        FORMAT_MODE = sys.argv[sys.argv.index('-f') + 1]
        if not FORMAT_MODE in VALID_FORMAT_MODES:
            print("Invalid format mode '" + FORMAT_MODE + "'")
            print("Valid modes are " + str(VALID_FORMAT_MODES)[1:-1])
            sys.exit(0)

    if '-s' in sys.argv:
        # sensitivity value
        try:
            SENSITIVITY_CONSTANT = float(sys.argv[sys.argv.index('-s') + 1])
        except ValueError:
            print("Sensitivity value must be numeric.")
            sys.exit(0)

    if '-p' in sys.argv:
        # ionization probability 
        try:
            ION_PROB_CONSTANT = float(sys.argv[sys.argv.index('-p') + 1])
        except ValueError:
            print("Ionization probability value must be numeric.")
            sys.exit(0)

    if '-g' in sys.argv:
        # frag factor
        try:
            FRAG_FACTOR_CONSTANT = float(sys.argv[sys.argv.index('-g') + 1])
        except ValueError:
            print("Frag factor value must be numeric.")
            sys.exit(0)

    write_output_file(sys.argv[1])
